/*
						 ***********
						** Summary **
						 ***********
	(1)	Title:
				Nagual Patch Utility

	(2)	Purpose:
			General purpose file haxor.

	(3)	Non-ANSI features:
			***  None  ***

	(4)	Platform dependencies:
			Recent GNU libc/libc++ i/o interfaces are b0rken wrt
			handling automatic open(2) via xfstream(...) constructors.
			Consequently, we perform our own open(2) with subsequent
			unification to an appropriate xfstream to restore sanity.

	(5)	Change history:
			Please consult the RCS logs for the history of changes.

	(6)	Input:

			Classes:
				***  None  ***

			Parameters:
				See input assertions

			Common:
				***  None  ***

			Files:
				See input assertions

			Assertions about input:

	Usage:	rochay  [-vta] [-o newFile]  fileToPatch  [patchCmdFile]  [script args]

	where -v causes rochay to recite its results from both the verification
	pass and the application pass.

	-t prevents rochay from actually applying any changes, inplies -v,
	and otherwise does everything except modify the file.

	-o specifies the name of an alternate output file to write the
	patch results to.  Imples no replacement of the original file.

	-a reverses the sense of the comparison command (?).

	fileToPatch is the name of the file whose contents are modified
	by rochay.  If the fileToPatch is supplied via the `|' command
	within the patch command file, use `-' as a place-keeper here.

	patchCmdFile, when present, specifies the name of the file containing
	the patch commands to apply to the fileToPatch.  If absent, rochay
	reads its commands from standard input, disables its first pass
	verification, and disables the insert/delete commands.

	[script args] may be any legal patch commands, but typically names
	a macro to invoke within the patch command file.

	Each command line token, beginning with [script args] is assigned
	to a script macro: %cliN, with N running from 1 to the total number
	of script arguments.

	In addition, the entire set of script arguments is raveled into a
	single string and assigned to the macro %cli.  So the specimen:

		rochay fall.exe items.pat %stamina %identify

	creates 3 macros for the patch script:

		"cli1
		%stamina
		"
		"cli2
		%identify
		"
		"cli
		%stamina %idenify
		"

	If the patch script contains the command "%cli1" the %stamina
	macro will be performed.  In this example, %cli isn't meaningful.

	General caveat: insert/delete operations will likely not work reliably
	                on network path names.  For rename() to work right,
					both files must reside on the same (local) system.

	Patch scripting language:
		See the seperate documentation: WereFall.html


	(7)	Output:

			Classes:
				***  None  ***

			Parameters:
				***  None  ***

			Common:
				***  None  ***

			Files:
				***  None  ***

			Assertions about the output:
				See the patch language assertions above.

	(8)	Public members:

		1.	Title:
					int main

			Purpose:
				Apply the patch script to the file name given on the CLI.

			Parameters:
				int argc			--	Count of cli tokens.
				char** argv			--	List of cli tokens.


		2.	Title:
					void pError

			Purpose:
				Gripe upon the current anomaly in the system.

			Parameters:
				const char* txt		--	Message text of the gripe d'jour.


	(9)	Private members:

		1.	Title:
					int cliConfig

			Purpose:
				Build the command line macros from an argv-style
				token list.

			Parameters:
				Cli& theCli			--	Configuration repository.
				int argc			--	Count of tokens to process.
				char** argv			--	List of tokens to process.
				return				--	Answer true.


		2.	Title:
					int cmdConfig

			Purpose:
				Build the command line macros from a raw command line.

			Parameters:
				Cli& theCli			--	Configuration repository.
				char* cmdLine		--	Command line to macrofy.
				return				--	Answer true.


		3.	Title:
					int copyChunk

			Purpose:
				Copy a chunk of inHandle to outHandle.  Answer false
				if an error was detected, and false otherwise.

			Parameters:
				int inHandle		--	The command input source.
				int outHandle		--	The copy output destination.
				long count			--	Number of bytes to copy.  May
										be < 1, which indicates no
										copy will be done.
				return				--	Answer false if an error was
										detected, and true otherwise. 


		4.	Title:
					int epilogue

			Purpose:
				To finalize the patch context for the current pair
				of original/modified files.

			Parameters:
				Cli& theCli			--	Current context.
				return				--	Answer 0 if all is as it should
										be, and an error indication
										otherwise.


		5.	Title:
					int patch

			Purpose:
				Process the patch script command found in cmdLine for
				application to the file associated with a handle.
				If applyChange is false, do everything except write
				out the updates.  If applyChange is true, commit the
				change to the file outHandle as well.

				Answer 0 if all is as it should be.  Else answer > 0
				if there is a syntax/file access error.  Answer -1 to
				skip the next record of the patch command file.  Answer
				-2 to print the next reocrd of the patch command file
				and then die.

			Parameters:
				Cli& theCli			--	The configuration repository.
				int outHandle		--	The command output destination.
				char* cmdLine		--	The patch command to digest.
				int applyChange		--	If true, commit the actual changes.
				return				--	Answer 0 if the command was processed
										successfully, and an error code
										otherwise.


		6.	Title:
					int prologue

			Purpose:
				To initialize the patch context for the current pair
				of original/modified files.

			Parameters:
				Cli& theCli			--	Current context.
				return				--	Answer 0 if all is as it should
										be, and an error indication
										otherwise.


		7.	Title:
					int readCmds

			Purpose:
				Process the contents of the patch script file.  If any
				patch attempt ends in error, stop processing the file
				and return immediately.

			Parameters:
				Cli& theCli			--	The configuration repository.
				int outHandle		--	The destination file for the
										patch command operation.
				istream* pFile		--	The patch command file source.
				int applyChange		--	If false, perform only the
										verification pass, else commit
										the actual updates as well.

		8.	Title:
					int replace
			Purpose:
				To scan the source buffer area for replaceable parts
				from the environment ($foo) or macro (%foo) tables.
				Answer true if replacement has occurred, and false
				otherwise.  The first character of the source buffer
				is always copied out verbatim.

			Parameters:
				char* src			--	Source buffer.
				char* dest			--	Destination buffer.
				int destSize		--	Size of destination area.
				Macro* set			--	Pool of current macro symbols.
				return				--	Answer true if any replacements
										occurred, and false otherwise.


		9.	Title:
					int usage

			Purpose:
				To emit a brief discription of the command use on
				standard output.

			Parameters:
				int errNo			--	Diagnostic return code.
				return				--	Answers errNo.


	(10) Subprograms called:
			***  None  ***

	(11) Includes referenced:
			***  None  ***

	(12) Classes declared:
			Cli						--	Patch operation context.
			InputStream				--	Read stream for file stream sets.
			Macro					--	A small macro collection.
			MacroStream				--	Read stream for Macro sets.
			StringList				--	A small string collection.

						 ****************
						** Declarations **
						 ****************

					 ***********************
					** Global declarations **
					 ***********************

				  Copyright 2001, Nagual Enterprises
			Please see COPYING for the complete GNU license

					  ********************
					 ** Global functions **
					  ********************
*/

#			ifndef __BORLANDC__
#				include <unistd.h>
#				define intBits 16
#			else
#				include <io.h>
#				define intBits 32
#			endif
#				include <stdlib.h>
#				include "file.h"
#				include <stdio.h>
#				include <sys/types.h>
#				include <sys/stat.h>
#				include <fcntl.h>
#				include <string.h>
#				include <ctype.h>
#				include <errno.h>

/*
					  *************
					 ** Constants **
					  *************
*/

					char WhoAmI[] = "rochay";
					char aCli[] = "cli";
					char aMac[] = "mac";
					const char copyRight[]=
						"rochay: Patch File Tool, version 1.1\n"
						"        Copyright (c) 2001, Nagual Enterprises";

/*
					   **********
					  ** Macros **
					   **********
*/
					/*	Helper macro to document integer conversion
						errors.

						Emit the system error message, the failing
						patch command line--marked by a carat at the
						point of failure--and the amplifying
						application error message. */

#				define ConvertError(msg, errNo) \
				{ \
					int position = 1 + (endPtr - cmd); \
					if (errno) \
						pError(WhoAmI); \
					cout << cmd << endl; \
					cout << setw(position) << setfill(' ') \
						 << "^" << endl << msg << endl; \
					return errNo; \
				}

					/*	Helper macro to document other parsing errors.

						Emit the failing patch command line--marked by
						a carat at the point of failure--and the
						amplifying application error message. */

#				define ParseError(msg, errNo) \
				{ \
					int position = 1 + (endPtr - cmd); \
					cout << cmd << endl; \
					cout << setw(position) << setfill(' ') \
						 << "^" << endl << msg << endl; \
					return errNo; \
				}

					/*	Helper macros to document system (i/o) errors.

						Emit the system error message, the failing
						patch command line, and the amplifying
						application error message. */

#				define SeekError(errNo) \
				{ \
					/*if (errno)*/ \
						pError(WhoAmI); \
					cout << cmd << endl; \
					cout << "Unable to scroll " << theCli.fileName \
						 << " to 0x" << hex << setw(8) << setfill('0') \
						 << offset << dec << " for patching." << endl; \
					return errNo; \
				}

#				define ReadError(errNo) \
				{ \
					if (errno) \
						pError(WhoAmI); \
					cout << cmd << endl \
						 << "Unable to read original value at offset 0x" \
						 << hex << setw(8) << setfill('0') << offset << dec \
						 << endl << "in '" << theCli.fileName << "'." << endl; \
					return errNo; \
				}

#				define WriteError(errNo) \
				{ \
					if (errno) \
						pError(WhoAmI); \
					cout << cmd << endl; \
					cout << "Unable to write patch value (0x" \
						 << hex << setw(8) << setfill('0') \
						 << ((unsigned int)*output) << ") at offset 0x" \
						 << setw(8) << setfill('0') << offset << dec; \
					cout << endl << "in '" \
						 << theCli.fileName << "'." << endl; \
					return errNo; \
				}

/*
					  ***********
					 ** Classes **
					  ***********
*/

					/*	Declare a small string collection class. */

				class StringList;
				class Macro;
				class StringList
				{
				protected:
					StringList* _next;			//	Next element in list.
					char* _text;				//	Contents of this element.

					void replace(Macro* set);	//	Symbolic substitution.

				public:
					StringList() :				//	Constructors, etc
						_next(NULL),
						_text(NULL) {}
					StringList(char* t) :
						_next(NULL),
						_text(t) {}
					StringList(	char* t,
								Macro* set) :
						_next(NULL),
						_text(t) {replace(set);}
					~StringList()
						{if (_text) free(_text);
						 if (_next) delete _next;}

					void add(StringList* m)		//	Add a string to the set.
						{_next ? _next->add(m) : (void)(_next = m);}
												//	Answer true if this is
					int atEnd()					//	the last element of the set.
						{return _next == NULL;}
					char* contents()			//	Answer this element's
						{return _text;}			//	string.
					void contents(char* c)		//	Assign this element's
						{_text = c;}			//	string.
					StringList* next()			//	Answer the next element in
						{return _next;}			//	this set.
				};

					/*	Declare a small collection of Macros. */

				class Macro
				{
				protected:
					Macro* _next;				//	Next element in the list.
					StringList* _lines;			//	Contents of this element.

				public:
					Macro() :					//	Constructors, etc
						_next(NULL),
						_lines(NULL) {}
					Macro(StringList* l) :
						_next(NULL),
						_lines(l) {}
					~Macro()
						{if (_lines) delete _lines;
						 if (_next) delete _next;}

					void add(Macro* m)			//	Add a line to this set.
						{_next ? _next->add(m) : (void)(_next = m);}
												//	Answer true if this is the
					int atEnd()					//	last element of the set.
						{return _next == NULL;}
					StringList* contents()		//	Answer the lines of this
						{return _lines;}		//	macro.
					void						//	Assign the lines of this
						contents(StringList* c)	//	macro.
						{_lines = c;}
					StringList*					//	Answer the lines of the
						find(const char* n,		//	macro with this name.
							 int len)
						{return isEqualI(n, name(), len) ? 
							(lastMatched = this, _lines) :
							(_next ?
								_next->find(n, len) :
								NULL);}
					void lines(StringList* l)
						{_lines = l;}
					char* name()
						{return _lines->contents();}
					Macro* next()
						{return _next;}
					void next(Macro* m)
						{_next = m;}
					void remove(Macro* m)
						{if (_next == m)
						 	(_next = m->next(), m->next(NULL));
						 else _next ?
						 	_next->remove(m) :
						 	(void)NULL;}

					static int isFrom;
					static int isIterate;
					static Macro* lastMatched;

					static StringList*
						findIn(	const char* n,
								Macro* list);
				};

					/*	Declare a read stream helper for record-
						oriented i/o.

						Warning: instance garbage collection at delete. */

				class InputStream;
				class MacroStream;
				class InputStream
				{
				protected:
					InputStream* _next;		//	Next (stacked) input stream.
					void* _baseStream;		//	The base collection.
					int _handle;			//	The fs handle.

				public:
					InputStream() :			//	Constructors, etc
						_next(NULL),
						_baseStream(NULL),
						_handle(-1) {}
					InputStream(istream* stream) :
						_next(NULL),
						_baseStream((void*)stream),
						_handle(-1) {}
#				ifndef __BORLANDC__
					InputStream(ifstream* stream) :
						_next(NULL),
						_baseStream((void*)stream),
						_handle(-1) {}
#				endif
					InputStream(istream* stream,
								int h) :
						_next(NULL),
						_baseStream(stream),
						_handle(h) {}
					virtual ~InputStream() {}

					virtual int atEnd()		//	line has already been used.
						{return ((istream*)_baseStream)->eof();}
					virtual void close()	//	Close the base collector.
						{if (_handle >= 0)
							::close(_handle);}
//						{((fstream*)_baseStream)->close();}
					istream* file()
						{return (istream*)_baseStream;}
					virtual int isError()
						{return ((istream*)_baseStream)->bad();}
					virtual void			//	Retrieve the next line.
						next(char* line, int lineSize)
						{if (!atEnd())
							((istream*)_baseStream)->getline(line, lineSize);
						else
							*line = '\0';}
					virtual int peek()		//	Peek at the next character.
						{return ((istream*)_baseStream)->peek();}
					InputStream* pop()
						{return _next;}
					virtual unsigned int	//	Answer the base stream
						position()			//	position
						{return ((istream*)_baseStream)->tellg();}
					virtual void			//	Reposition the base stream.
						position(unsigned int i)
						{((istream*)_baseStream)->seekg(i - 1, ios::beg);}
					void push(InputStream* s)
						{_next ? _next->push(s) : (void)(_next = s);}
											//	Answer true if the last
					virtual void reset()	//	Rewind the base stream.
						{((istream*)_baseStream)->seekg(0, ios::beg);}
					virtual void skip(int i)//	Skip over lines.
						{((istream*)_baseStream)->seekg(i, ios::cur);}
				};

					/*	Subclass InputStream for macro collections. */

				class MacroStream : public InputStream
				{
				protected:
					StringList* _list;		//	The current position.
					void advance()			//	Advance to next line.
						{if (_list)
							_list = _list->next();}
				public:						//	Constructors, etc
					MacroStream() :
						InputStream(),
						_list(NULL) {}
					MacroStream(istream* stream) :
						InputStream(stream),
						_list(NULL) {}
					MacroStream(StringList* m) :
						InputStream(),
						_list(m) {_baseStream = (void*)m;}
					virtual ~MacroStream() {}

											//	Answer true if the last line
					virtual int atEnd()		//	has already been used.
						{return _list == NULL;}
					virtual void close() {}	//	Close the base stream.
					virtual int isError()
						{return false;}
					virtual void			//	Answer the next line.
						next(char* line, int lineSize)
						{if (!atEnd())
						{
							memccpy(line, _list->contents(), '\0', lineSize);
							advance();
						}
						else
							*line = '\0';}
					virtual int peek()		//	Peek at the next character.
						{return *_list->contents();}
					virtual void reset()	//	Rewind the macro.
						{_list = (StringList*)_baseStream;}
					virtual void skip(int i)//	Skip over lines.
						{if (!atEnd())
							while (i--) advance();}
				};

					/*	Collect the command configuration into a common
						pool. */

				struct _Cli
				{
					int handle;				//	Handle on original file.
					int modHandle;			//	Handle on copied file.
					ifstream* patch;		//	Stream on patch commands.
					int apply;				//	True to apply changes.
					int allowCopy;			//	True to copy changes over
											//	the original.
					char* copyName;			//	File name for copy-replace.
					long originalSize;		//	File size of the original.
					char* fileName;			//	Original file name.
					char* patchName;		//	Patch command script.
					int verify;				//	True to report patch progress.
					int needCopy;			//	True if in-place mods won't work.
					int cantCopy;			//	True if verification pass failed.
					long lastOffset;		//	Last offset accessed in original.
					long outputBias;		//	Output bias relative to original,
											//	accounting for inserts/deletes.
					int lastCmd;			//	Last patch command processed.
					long dot;				//	The `dot' position in original.
					long tableIncr;			//	The table offset increment.
					int incrSize;			//	The byte size of the table offset.
					int macIterate;			//	Current macro iteration.
					Macro* macros;			//	Current pool of macros.
					int patchPass;
					int reply;				//	Latest error indication.
					int haveRedirect;
					int allowCreate;		//	True to create-on-open.
					int idleHandle;
					char* incFile;			//	Byte stream include file name.
					int invert;				//	True to invert (?) operation.
				};
				typedef struct _Cli Cli;

/*
					   **********
					  ** Common **
					   **********
*/

					long byteMasks[]={
						0, 0xff, 0xffff, 0xffffff, 0xffffffff};
					Macro* Macro::lastMatched = NULL;
					int Macro::isFrom = 0;
					int Macro::isIterate = 0;
					char copyName[256];

/*
					 **************
					** Parameters **
					 **************


				 **********************
				** Local declarations **
				 **********************
*/

					static int cliConfig(Cli& theCli,
										 int argc,
										 char** argv);
					static int copyChunk(int inHandle,
										 int outHandle,
										 long chunk);
					static int epilogue(Cli& theCli);
					static int patch(Cli& theCli,
									 int outHandle,
									 char* cmdLine,
									 int applyChange);
					static int prologue(Cli& theCli);
					static int readCmds(Cli& theCli,
										int outHandle,
										int applyChange);
					static int replace(	char* src,
										char* dest,
										int destSize,
										Macro* set);
					static int usage(int errNo);
					void pError(const char* txt);

/*
					  *************
					 ** Variables **
					  *************
*/

					static int needCopy[32]={
						0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0
					};
					static int section = 0;
					static char fileName[256];

/*
						 ********
						** Text **
						 ********


							 *******************
							** Private methods **
							 *******************
*/

static char* stros1
(	char* src)
{
		/*	Transform a path name in its own OS dialect. */

	char* s = src;
	while (*s)
		if (*s == '/')
			*s++ = DirSeperator;
		else
			*s++;
	return src;
}

static int cliConfig
(	Cli& theCli,
	int argc,
	char** argv)
{
		/*	Build the command line macros. */

	StringList* cli = new StringList(aCli);
	int cliNum = 1;
	StringList* macro = NULL;
//	if (argc)
//	{
		if (!theCli.macros)
			/*	Anchor the list of macros. */

			theCli.macros = new Macro(cli);
		else
			theCli.macros->add(new Macro(cli));
//	}
	while (argc--)
	{
		char* text = (char*)malloc(16);
		char buff[16];
		ostrstream tag(buff, 16);
		tag << aCli << cliNum << ends;
		strcpy(text, buff);
		Macro* value = new Macro(new StringList(text));
		theCli.macros->add(value);
		value->contents()->add(new StringList(strdup(*argv)));
		cli->add(new StringList(strdup(*argv)));
		text = (char*)malloc(16);
		char macBuff[16];
		ostrstream mtag(macBuff, 16);
		mtag << aMac << cliNum++ << ends;
		strcpy(text, macBuff);
		value = new Macro(new StringList(text));
		theCli.macros->add(value);
		value->contents()->add(new StringList(strdup(*argv++)));
	}
	return true;
}

static int cmdConfig
(	Cli& theCli,
	char* cmdLine)
{
		/*	Build the command parameters. */

	int cliNum = 0;
	char* next = cmdLine;

	while (*next)
	{
		char* start = next;
		while (!isspace(*start) && *start)
			start++;

		/*	Ignore the invocation pattern. */

		if (cliNum)
		{
			char* text = (char*)malloc(16);
			char buff[16];
			ostrstream tag(buff, 16);
			tag << aMac << cliNum++ << ends;
			strcpy(text, buff);

			/*	Create %mac definition. */

			Macro* value = new Macro(new StringList(text));

			/*	If %macN already exists, garbage collect its prior
				value(s). */

			StringList* oldMac = Macro::findIn(text, theCli.macros);
			if (oldMac)
			{
				theCli.macros->remove(Macro::lastMatched);
				delete Macro::lastMatched;
				Macro::lastMatched = NULL;
			}
			/*	Copy off the new value for the symbol. */

			text = (char*)malloc(start - next);
			char* cPtr = next;
			char* dPtr = text;
			while (cPtr != start)
				*dPtr++ = *cPtr++;
			*dPtr = '\0';

			/*	Add it to the definition. */

			value->contents()->add(new StringList(text));
			theCli.macros->add(value);
		}
		else
		{
			cliNum++;
			*start++ = '\0';
		}
		while (isspace(*start) && *start)
			start++;
		next = start;
	}
	return true;
}

static int copyChunk
(	int inHandle,
	int outHandle,
	long chunk)
{
		/*	Copy a chunk of inHandle to outHandle.  Answer true if no
			errors were detected, and false otherwise.  Use the present
			positions of the handles to begin operations at. */

	if (chunk > 0)
	{
		char buffer[8192];
		size_t size;
		while (chunk > 0)
		{
			size = chunk > sizeof(buffer) ?
				sizeof(buffer) :
				chunk;
			errno = 0;
			if (read(inHandle, buffer, size) != size)
			{
				pError("read(2)");
				return false;
			}
			errno = 0;
			if (write(outHandle, buffer, size) != size)
			{
				pError("write(2)");
				return false;
			}
			chunk -= size;
		}
	}
	return true;
}

static int epilogue
(	Cli& theCli)
{
		/*	If the patch requires insert/delete operations, create
			a new file to receive the result of the edits. */

	/*	If the syntax scan passes or patching from stdin, apply the
		requisite changes to the file. */

	if (::close(theCli.handle))
		cout << "--close handle did not answer 0." << endl;
	if (theCli.handle != theCli.modHandle)
	{
		if (::close(theCli.modHandle))
			cout << "--close modhandle did not answer 0." << endl;
	}
	if (!theCli.patchPass && theCli.needCopy)
	{
		/*	Record the need for copy for future reference. */

		needCopy[section >> intBits] |= 1 << (section & (intBits - 1));
	}
	if (!theCli.reply && theCli.patchPass)
	{
		if (theCli.handle != theCli.modHandle && theCli.allowCopy)
		{
			/*	If insert/delete was used, rename the editted file unless the
				edit sequence has spontaneously failed for some reason,
				or we're operating in `test' mode. */

			if (theCli.needCopy && !theCli.reply && theCli.apply && theCli.allowCopy)
			{
				errno = 0;
//				if (rename(theCli.copyName, theCli.fileName))
//				{
					/*	Hmmm, this shouldn't happen in practice, (he sezs
						theoretically).  But just in case someone fools
						around with the user's access privileges while he
						isn't looking...or *gack* uses the original directory
						as a mount point... */

//					pError(WhoAmI);
//					cout
//						<< "Replacing '"
//						<< theCli.fileName
//						<< "' with its new contents has failed.\n"
//						<< "The editted contents may be found in '"
//						<< theCli.copyName
//						<< "' until the tool is next used."
//						<< endl;
//					theCli.reply = errno;
//				}
				theCli.copyName = NULL;
			}
		}
	}
	theCli.allowCreate = false;
	return theCli.reply;
}

static int patch
(	Cli& theCli,
	int outHandle,
	char* cmdLine,
	int applyChange)
{
		/*	Process the patch script command found in cmdLine for application
			to the file associated with a handle.  If applyChange is false,
			do everything except write out the updates.  If applyChange is
			true, commit the change to the file outHandle as well.

			Answer 0 if all is as it should be.  Else answer > 0 if there
			is a syntax/file access error.  Answer -1 to skip the next record
			of the patch command file.  Answer -2 to print the next record
			of the patch command file and then die. */

	long offset, value, original, master;
	long oldBias;
	char* endPtr;
	char* cmd = cmdLine;
	int multiplier = 1;
	int dontReplace;
	char* output;
	char* outBuffer;
	int reply = 0;
	int haveInsert = 0;
	int haveDelete = 0;
	int haveOverlay = 0;
	int haveDot = 0;
	int isRelative = 0;
	int haveIncr = 0;
	int size = 1;
	int branch;
	int isCont = 0;
	int byteStream = 0;

	/*	Get the next file offset and sanitize it to death. */

	while (*cmdLine && isspace(*cmdLine))
		cmdLine++;
	if (!*cmdLine)
		return 0;
	isRelative = *cmdLine == '+' || *cmdLine == '-';
	errno = 0;
	offset = strtol(cmdLine, &endPtr, 0);
	if (errno)
		ConvertError("Unable to convert offset.", 1);

	/*	Check for verify, insert, or delete modifiers. */

	dontReplace = *endPtr == '?' && offset == 0;
	if (offset == 0)
	{
		isCont = endPtr[1] == '+';
		if (*endPtr == '>')
			theCli.needCopy = haveInsert = 4;
		else if (*endPtr == '<')
			theCli.needCopy = haveDelete = 8;
		else if (*endPtr == '@')
		{
			oldBias = theCli.outputBias;

			haveOverlay = 32;
		}
		else if (*endPtr == '.')
		{
			if (!endPtr[1])
			{
				theCli.dot = theCli.lastOffset;
				return 0;
			}
			haveDot = 1;
		}
		else if (*endPtr == '=')
		{
			haveIncr = 16;
			isRelative |= isCont;
			isCont = false;
		}
		else if (*endPtr == '#')
		{
			cout
				<< "-- output bias is 0x"
				<< hex << setfill('0')
				<< theCli.outputBias
				<< dec << '.'
				<< endl;
			return 0;
		}
		else if (dontReplace)
		{
			isRelative = 1;
			isCont = false;
		}
	}
	/*	Select different cases according to which pass of the
		patch we're presently performing. */

	branch = haveInsert |
			 haveDelete |
			 haveOverlay |
			 haveIncr |
			 ((applyChange & 1) << 1) |
			 (dontReplace & 1);

	/*	Whine and die if his need for copy can't be obliged. */

	if (theCli.needCopy && theCli.cantCopy)
		;
//		ParseError("Cannot use insert/delete with piped commands.", 11);

	/*	Replacing parts following an insert or delete isn't the brightest
		idea in the world as it's far too easy to confuse what the proper
		offset(s) now are.  Refuse to humor this idiom for absolute
		addressing. */

	if (*cmdLine && theCli.needCopy && (branch > 18 || branch < 4) &&
		(!haveDot && !dontReplace))
		ParseError(	"Replace following insert/delete is error-prone.\n"
					"Please reorder the patch commands.", 12);

	/*	If we found an address modifier, fetch the address it operates
		on. */

	isRelative |= dontReplace && (endPtr[1] == '+' || endPtr[1] == '-');
	if (dontReplace || haveInsert || haveDelete || haveOverlay || haveDot ||
		haveIncr)
	{
		/*	Get the file offset to begin verification at.  Mark the
			`skip next record' return code for the caller. */

		if (!haveOverlay && branch < 4)
			reply = -1;
		cmdLine = &endPtr[1];
		if (*cmdLine == '^')
		{
			cmdLine++;
			errno = 0;
			int hand = ::open(cmdLine, O_RDONLY);
			if (hand < 0)
			{
				pError(WhoAmI);
				cout << "Unable to include bytes from "
					<< *cmdLine
					<< "'."
					<< endl;
				return 24;
			}
			theCli.incFile = theCli.fileName;
			theCli.idleHandle = theCli.handle;
			theCli.handle = hand;
			theCli.fileName = cmdLine;
			byteStream = true;
			offset = theCli.dot + lseek(hand, 0, SEEK_END);
		}
		else
		{
			offset = strtol(cmdLine, &endPtr, 0);
			if (errno)
				ConvertError("Unable to convert offset.", 1);
		}
	}
	if (haveIncr && !isRelative)
	{
		switch (*endPtr)
		{
			case 'l':
				theCli.incrSize = 4;
				break;
			case 's':
				theCli.incrSize = 2;
				break;
			case 'm':
				theCli.incrSize = 3;
				break;
			default:
				theCli.incrSize = 1;
		}
		theCli.tableIncr = offset;
		return 0;
	}
	/*	Check for continuation from previous offset. */

	isCont |= offset == 0 && *endPtr == '+';
	if (isCont)
	{
		offset = theCli.lastOffset;
//		if (haveInsert)
//			offset++;
	}
	else if (isRelative && !haveIncr)
		offset += theCli.dot;
	else if (*endPtr == '>')
		offset = theCli.originalSize;

	/*	Check the requested offset exists within the file. */

	if (!dontReplace && !haveInsert && offset > theCli.originalSize)
		ParseError("File is smaller than requested offset.", 10);

	if (haveDot)
	{
		theCli.dot = offset;
		return 0;
	}
	/*	In fact, since we don't sort the insert/delete offsets,
		detect whether he has ignored the canonical increasing
		offset for such edits.  Whine and die, if so. */

	if (offset < theCli.lastOffset && (theCli.lastCmd & 12))
		ParseError("Offset smaller than previous offset.\n"
					"Please order edit in increasing order of offsets.", 13);

	/*	Okay, this offset might do.  Advance to the data specification.
		Iterate over all its components. */

	cmdLine = &endPtr[1];
	while (*cmdLine && isspace(*cmdLine))
		cmdLine++;
	while (*cmdLine)
	{
		int isQuoted;
		int iterations = 1;
		long outOff;

		/*	Just escape at line end. */

		if (*cmdLine == '\n' || *cmdLine == '\r' || !*cmdLine)
			break;

		/*	Try to convert this token as a basic integer value. */

		value = strtol(cmdLine, &endPtr, 0);
		if (errno)
			ConvertError("Unable to convert data value.", 2);
		if (haveIncr)
		{
			offset = *cmdLine == '+' ?
				theCli.dot + value :
				value;
			size = theCli.incrSize;
		}
		else switch (*endPtr)
		{
			/*	Adjust the atomic size of the data value. */

			case 's':
				size = 2;
				break;
			case 'm':
				size = 3;
				break;
			case 'l':
				size = 4;
				break;
			default:
				size = 1;
				break;
		}
		if (*endPtr == ':')
		{
			/*	Translate a texture picture code. */

			cmdLine = &endPtr[1];
			int index =strtol(cmdLine, &endPtr, 0);
			if (errno)
				ConvertError("Unable to convert picture index.", 2);
			value = (value << 7) + (index & 0x7f);
			size = 2;
		}
		/*	Look for intervening repetition marks (*) */

		if (endPtr[1] == '*')
			endPtr++;
		if (*endPtr == '*')
		{
			/*	Use the previous value for the repetition count and
				fetch the next value to replicate. */

			multiplier = value;
			cmdLine = &endPtr[1];
			value = strtol(cmdLine, &endPtr, 0);
		}
		if (haveOverlay)
		{
			theCli.outputBias = offset - value;
			offset = value;
			cmdLine = &endPtr[1];
			value = strtol(cmdLine, &endPtr, 0);
			cmdLine = &endPtr[1];
		}
		/*	Check if the specified value is a string, rather than an
			integer.  (Just throw away `value' if so)  Ensure the
			quotes are balanced. */

		isQuoted = *endPtr == '"' || *endPtr == '\'';
		if (value && isQuoted)
			ParseError("Didn't understand number followed by quote mark.", 6);
		if (isQuoted)
		{
			/*	Since we've got an opening quote, churn through the letters,
				looking for its closing mate. */

			char quoted = *endPtr;
			char* endQuote;
			cmdLine = &endPtr[1];

			endQuote = strchr(cmdLine, quoted);

			/*	Whine if we didn't locate one. */

			if (!endQuote)
				ParseError("Quoted string does not end with quote mark.", 7);

			/*	Figure the number of characters in the string. */

			iterations = endQuote - cmdLine;
			endPtr = endQuote;
			output = cmdLine;
		}
		else
		{
			if (haveDelete || haveOverlay)
			{
				iterations = value;
				output = (char*)&original;
			}
			else if (haveIncr)
				output = (char*)&original;
			else
				/*	Otherwise, just handle a single byte value. */

				output = (char*)&value;
		}
		/*	Confirm that some useful work was actually located. */

		if (cmdLine == endPtr)
			ParseError("Didn't find a data value to convert.", 8);

		outBuffer = output;
		if ((theCli.needCopy && applyChange && !dontReplace) ||
			theCli.handle != outHandle)
		{
			/*	Copy any intervening bytes from the end of the last edit
				to the start of the present one. */

			long outSize = offset - theCli.lastOffset;
			int needFill = true;
			if (haveOverlay)
			{
				outSize += theCli.outputBias;
			}
			long cur = lseek(outHandle, 0, SEEK_CUR);
			long fin = lseek(outHandle, 0, SEEK_END);
			lseek(outHandle, cur, SEEK_SET);
			needFill = cur == fin;
			if (needFill)
				if (!copyChunk(theCli.handle, outHandle, outSize))
					WriteError(5);
		}
		if (!byteStream)
		while (multiplier--)
		{
			int count = iterations;
			while (count--)
			{
				/*	Check the requested offset exists within the file. */

				if (!haveDelete && !haveInsert && offset > theCli.originalSize)
					ParseError("File is smaller than requested offset.", 10);

				/*	Scroll to the next byte to process. */

				errno = 0;
				if (lseek(theCli.handle, offset, SEEK_SET) != offset)
					SeekError(3);

				/*	Actually read in the next byte. */

				errno = 0;
				original = 0;
				if (!haveInsert && !haveDelete)
					if ((read(theCli.handle, &original, size) != size) || errno)
						ReadError(4);
				master = original;
				switch (branch)
				{
					case 1:
					case 3:
						/*	If dontReplace (e.g. verify contents) then just
							compare the original with the proposed change.
							If they don't match, return code -2 for the
							caller's benefit. */

						if (theCli.invert)
						{
							if (original != *output)
								break;
						}
						else
						{
							if (original == *output)
								break;
						}
						return -2;

					case 16:
						original += theCli.tableIncr;
						break;

					case 18:
						/*	Apply the table increment to this slot, and
							fall through to output. */

						original += theCli.tableIncr;

					case 34:
						/*	Overlay bytes from the input stream. */
					case 6:
						/*	Insert new bytes in the output stream. */

					case 2:
						/*	if (applyChange) commit the verified changes
							on the second pass. */

						outOff = offset + theCli.outputBias;
						errno = 0;
						if (outOff >= 0)
							if (lseek(outHandle, outOff, SEEK_SET) != outOff)
								SeekError(23);
						if ((write(outHandle, output, size) != size) || errno)
							WriteError(25);
						break;

					case 10:
						/*	Delete bytes from the output stream. */

						theCli.outputBias--;
						break;

					default:
						/*	Default is do nothing. */

						break;
				}
				if (branch == 6 || branch == 4)
					theCli.outputBias += size;
				if (theCli.verify)
				{
					/*	Document the change if requested to do so. */

					long* fin = (long*)output;
					cout
						<< "- 0x"
						<< hex
						<< (offset + theCli.outputBias)
						<< ": 0x"
						<< hex << setw(size) << setfill('0')
						<< master
						<< " --> 0x"
						<< hex << setw(size) << setfill('0')
						<< (*fin & byteMasks[size])
						<< '.'
						<< dec << endl;
				}
				if (!haveInsert)
					offset += size;
				if (!haveOverlay)
					output += size;
			}
			/*	Reset the buffers for the next iteration. */

			output = outBuffer;
		}
		/*	Finally, detect where in the original file we've ended up
			and advance to the next portion of the data specification. */

		multiplier = 1;
		if (!dontReplace)
		{
			if (haveInsert && !applyChange)
				theCli.lastOffset = offset;
			else
				theCli.lastOffset = lseek(theCli.handle, 0, SEEK_CUR);
		}
		else
			theCli.lastOffset = lseek(theCli.handle, 0, SEEK_CUR);
		cmdLine = &endPtr[1];
	}
	if (haveOverlay)
	{
		theCli.lastOffset = lseek(theCli.modHandle, 0, SEEK_END);
		theCli.outputBias = oldBias;
		lseek(theCli.handle, theCli.lastOffset, SEEK_SET);
	}
	theCli.lastCmd = branch;
	return reply;
}

static int prologue
(	Cli& theCli)
{
		/*	Open the file to operate on. */

	theCli.dot = theCli.lastOffset = theCli.outputBias = 0;
	if (!theCli.fileName)
		return 0;
	theCli.modHandle = errno = 0;
	int handle = ::open(theCli.fileName,
						(theCli.needCopy	? binaryI : binaryO) |
						(theCli.allowCreate	? O_CREAT : 0));
	theCli.handle = handle;
	if (theCli.handle < 1 || errno)
	{
		pError(WhoAmI);
		cout
			<< "Unable to open '"
			<< theCli.fileName
			<< "' for patching."
			<< endl;
		theCli.reply = 3;
		return usage(3);
	}
	int notInPlace = !theCli.copyName;
	if (notInPlace)
	{
		/*	Make up a suitable copy-replacement file name. */

		int len = strlen(theCli.fileName) + 1;
		if (len > sizeof(copyName))
		{
			cout
				<< theCli.fileName
				<< endl
				<< "File name longer than 255 characters."
				<< endl
				<< "Cannot create copy destination file."
				<< endl;
			theCli.reply = 17;
			return 17;
		}
		strncpy(copyName, theCli.fileName, len);
		char* base = strrchr(copyName, DirSeperator);
		if (!base)
		{
			/*	Cheap check for drive letter in Dos-land. */

			if (copyName[1] == ':')
				copyName[2] = '@';
			else
				*copyName = '@';
		}
		else
			base[1] = '@';
		theCli.copyName = copyName;
	}
//	else
//		theCli.allowCopy = false;
	cout
		<< "-- Running '"
		<< theCli.patchName
		<< "' on '"
		<< theCli.fileName
		<< '\''
		<< endl;
	if (theCli.patchPass)
	{
		if (!notInPlace)
			cout
				<< "-- Updating it in '"
				<< theCli.copyName
				<< "'."
				<< endl;
		else
			cout
				<< "-- No destination copy requested."
				<< endl;
	}
	/*	Get the original file size. */

	theCli.originalSize = lseek(theCli.handle, 0, SEEK_END);
	lseek(theCli.handle, 0, SEEK_SET);
	cout
		<< "-- Original size is "
		<< theCli.originalSize
		<< '.'
		<< endl;

	if ((!notInPlace || theCli.needCopy) && !theCli.reply && theCli.patchPass)
	{
		/*	Open a handle on the copy-replacement file. */

		struct stat statBuf;
		fstat(theCli.handle, &statBuf);
		theCli.modHandle = ::open(	theCli.copyName,
									binaryO | O_CREAT | O_TRUNC,
									statBuf.st_mode);
		if (theCli.modHandle < 0)
		{
			pError(WhoAmI);
			cout
				<< "Unable to open scratch file for insert/delete editing."
				<< endl;
			theCli.reply = 3;
			return usage(3);
		}
	}
	else
	{
		if (theCli.patchPass)
			theCli.cantCopy = true;
		theCli.modHandle = theCli.handle;
	}
	/*	If patching from a command file, process the file once to
		verify the syntax, but don't actually apply any changes. */

	return theCli.reply = 0;
}

static int readCmds
(	Cli& theCli,
	int outHandle,
	int applyChange)
{
		/*	Process the contents of the patch script file.  If any patch
			attempt ends in error, stop processing the file and return
			immediately. */

	int reply = 0;
	int inMacro = 0;
	int macTimes = 0;
	int macVer = theCli.verify;
	int macNoVer = 0;
	StringList* macro = NULL;
	InputStream* inStream = new InputStream(theCli.patch);
	InputStream* input = inStream;

	do
	{
		while (!input->atEnd())
		{

			/*	Retrieve the next line of the script. */

			char rawBuffer[1024];
			char repBuffer[1024];
			char* cmdBuffer = rawBuffer;
			memset(cmdBuffer, '\0', sizeof(rawBuffer));
			input->next(cmdBuffer, sizeof(rawBuffer));
			if (input->isError())
				break;
			/*	Echo the contents of the patch file in verify mode or if
				actually applying changes.  Special-case the contingent
				?0x... verification pattern to suppress the echo of the
				subsequent error message record unless verification
				is turned on.

				Note that here we're testing the reply from the *previous*
				script line. */

			if (!macNoVer)
				if ((reply >= 0) && (applyChange || theCli.verify))
					cout
						<< cmdBuffer
						<< endl;
			if (::replace(rawBuffer, repBuffer, sizeof(rawBuffer), theCli.macros))
				cmdBuffer = repBuffer;
			if (*cmdBuffer == '"')
			{
				if (applyChange)
				{
					/*	Skip forward past the end-of-macro. */

					do
						input->next(cmdBuffer, sizeof(rawBuffer));
					while (*cmdBuffer != '"');
					continue;
				}
				if (inMacro)
				{
					inMacro = 0;
					macro = NULL;
				}
				else
				{
					/*	Start a new macro definition. */

					inMacro = 1;
					macro = new StringList(strdup(&cmdBuffer[1]));
					StringList* oldMac = Macro::findIn(&cmdBuffer[1], theCli.macros);
					if (oldMac)
					{
						cout
							<< "-- Redefining macro."
							<< endl;
						theCli.macros->remove(Macro::lastMatched);
						delete Macro::lastMatched;
						Macro::lastMatched = NULL;
					}
					theCli.macros->add(new Macro(macro));
				}
				continue;
			}
			if (inMacro)
			{
				/*	Add this command to the current macro. */

				macro->add(new StringList(strdup(cmdBuffer)));
				continue;
			}
			if (*cmdBuffer == '|' || *cmdBuffer == '!')
			{
				/*	Change the file set the script operates on. */

				if (theCli.haveRedirect)
				{
					/*	This marks the end of the current patch section. */

					/*	Finalize this section. */

					if (theCli.needCopy && applyChange ||
						theCli.handle != outHandle)
						/*	Copy any remaining bytes left at the tail
							of the input file. */

						theCli.reply |= !copyChunk(	theCli.handle,
													outHandle,
													theCli.originalSize - theCli.lastOffset);

					if (epilogue(theCli))
						return theCli.reply;
				}
				/*	Start a new patch file section. */

				if (theCli.haveRedirect)
					section++;
				theCli.haveRedirect = true;
				theCli.lastOffset = 0;
				char* dlm = strchr(&cmdBuffer[1], *cmdBuffer);
				if (!dlm)
				{
					cout
						<< cmdBuffer
						<< endl
						<< " ^"
						<< endl
						<< "Second bar ("
						<< *cmdBuffer
						<< ") is missing."
						<< endl;
					return 19;
				}
				stros1(&cmdBuffer[1]);
				*dlm++ = '\0';
				theCli.allowCreate = *cmdBuffer == '!';
				if (theCli.patchPass)
				{
					/*	Retrieve the archived needCopy bit. */

					theCli.needCopy = needCopy[section >> intBits] &
										(1 << (section & (intBits - 1)));
				}
				else
					theCli.needCopy = false;
				theCli.fileName = strcpy(fileName, &cmdBuffer[1]);
				theCli.copyName = *dlm ?
					strcpy(copyName, dlm) :
					NULL;
				if (prologue(theCli))
					return theCli.reply;
				outHandle = theCli.modHandle;
				continue;
			}
			if (*cmdBuffer == '%')
			{
				/*	Try to invoke the named macro. */

				cmdConfig(theCli, cmdBuffer);
				StringList* mac = Macro::findIn(&cmdBuffer[1], theCli.macros);
				if (!mac)
				{
					cout
						<< cmdBuffer
						<< endl
						<< " ^"
						<< endl
						<< "Macro name could not be found."
						<< endl;
					return 14;
				}
				/*	Push the macro onto the input stack. */

				if (Macro::isIterate)
				{
					/*	Grab the iteration count from the invocation line. */

					if (theCli.incrSize < 1)
					{
						cout
							<< cmdBuffer
							<< endl
							<< " ^"
							<< endl
							<< "Table increment must be > 0."
							<< endl;
						return 17;
					}
					int count = strtol(&cmdBuffer[9], NULL, 0);
					if (count < 1)
					{
						cout
							<< "-- %iterate macro count < 1.\n-- Ignored."
							<< endl;
						continue;
					}
					if (theCli.macIterate)
					{
						cout
							<< cmdBuffer
							<< endl
							<< " ^"
							<< endl
							<< "Iteration macro already active."
							<< endl;
						return 15;
					}
					theCli.macIterate = count;
				}
				else if (Macro::isFrom)
				{
					/*	Grab the iteration count from the invocation line. */

					if (theCli.incrSize < 1)
					{
						cout
							<< cmdBuffer
							<< endl
							<< " ^"
							<< endl
							<< "Table increment must be > 0."
							<< endl;
						return 17;
					}
					char* endPtr;
					int start = strtol(&cmdBuffer[6], &endPtr, 0);
					if (!*endPtr || endPtr == &cmdBuffer[5])
					{
						cout
							<< cmdBuffer
							<< endl
							<< " ^"
							<< endl
							<< "%from must specify both -startAt- "
								"and -endAt- offsets."
							<< endl;
						return 16;
					}
					char* cPtr = endPtr;
					int finish = strtol(cPtr, &endPtr, 0);
					int count;
					count = ((finish - start) + theCli.tableIncr - 1) /
							theCli.tableIncr;
					if (count < 1)
					{
						cout
							<< "-- %from macro count < 1.\n-- Ignored."
							<< endl;
						continue;
					}
					if (theCli.macIterate)
					{
						cout
							<< cmdBuffer
							<< endl
							<< " ^"
							<< endl
							<< "Iteration macro already active."
							<< endl;
						return 15;
					}
					cout
						<< "-- Perform %from "
						<< count
						<< " times."
						<< endl;
					theCli.macIterate = count;
					macTimes = 0;
				}
				InputStream* newStream;
				newStream = (InputStream*)new MacroStream(mac);
				newStream->skip(1);
				newStream->push(input);
				input = newStream;
				continue;
			}
			if (*cmdBuffer == '^')
			{
				/*	Push a file onto the input stack. */

				cmdConfig(theCli, cmdBuffer);
				errno = 0;

				/*	Opening the file for the stream class ourselves seems
					to avoid more GNU libc brokenness.  Humor it. */
 
				int incHandle = ::open(&cmdBuffer[1], O_RDONLY);
				if (incHandle < 0)
				{
					pError(WhoAmI);
					cout
						<< cmdBuffer
						<< endl
						<< " ^"
						<< endl
						<< "Cannot include file '"
						<< "'."
						<< endl;
					return 14;
				}
				ifstream* include = new ifstream(incHandle);

				/*	Push the file onto the input stack. */

				InputStream* newStream = new InputStream(include, incHandle);
				newStream->push(input);
				input = newStream;
				continue;
			}
			/*	Just advance on comment lines. */

			if (*cmdBuffer == '-')
				continue;

			/*	If the verification test case succeeded, echo the error
				message as a comment, but only in [-v]erification mode.
				(Too confusing otherwise, to emit error message text
				unless there is truly an error) */

			if (theCli.verify && !applyChange && reply == -1)
				cout
					<< "- (else "
					<< cmdBuffer
					<< endl;
			if (reply == -1)
			{
				/*	Skip the error message line when the verification check
					passed. */

				reply = 0;
				continue;
			}
			if (reply == -2)
			{
				/*	Emit the verification error message when the check has
					failed. */

				cout
					<< "Verify failed:\n  "
					<< cmdBuffer
					<< endl;
				reply = 0;
				return 9;
			}
			/*	Finally, digest the current patch command.  Escape if it
				fails to parse or otherwise behave right. */

			reply = patch(theCli, outHandle, cmdBuffer, applyChange);
			if (reply > 0)
				return reply;
		}
		if (theCli.macIterate > 0)
		{
			/*	See if this macro iteration is complete. */


			if (macTimes++ > 4 && (theCli.verify || !macNoVer) && theCli.macIterate > 4)
			{
				macNoVer = true;
				theCli.verify = false;
				cout
					<< "-- ..."
					<< endl;
			}
			else if (theCli.macIterate < 4)
			{
				macNoVer = false;
				theCli.verify = macVer;
			}
			if (--theCli.macIterate > 0)
			{
				/*	Yet another cycle to go. */

				input->reset();
				input->skip(1);
				continue;
			}
			theCli.verify = macVer;
		}
		/*	Pop the input stack.  Escape if the stack is empty. */

		InputStream* top = input->pop();
		if (input != inStream)
			input->close();
		delete input;
//		if (!top)
//			break;
		input = top;
	} while (input);
	if (theCli.needCopy && applyChange ||
		theCli.handle != outHandle)
		/*	Copy any remaining bytes left at the tail of the input file. */

		reply |= !copyChunk(theCli.handle,
							outHandle,
							theCli.originalSize - theCli.lastOffset);

	return reply;
}

static int replace
(	char* src,
	char* dest,
	int destSize,
	Macro* set)
{
		/*	Replace the contents of the receiver with symbolic values
			from a macro set.  Answer true if replacement occurred, and
			false otherwise. */

	char* ptr = &src[1];
	char* bPtr = &dest[1];
	int runOn = destSize;
	int mustReplace = false;

	*dest = *src;
	while (*ptr && runOn--)
	{
		if ((*ptr == '$' || *ptr == '%') && !isspace(ptr[1]))
		{
			/*	Extract this symbol name off into its own quiet little
				corner. */

			char* ePtr = &ptr[1];
			while (!ispunct(*ePtr) && *ePtr)
				ePtr++;
			int size = (unsigned long)ePtr - (unsigned long)ptr;
#		ifdef __BORLANDC__
			char symbol[512];
#		else
			char symbol[size];
#		endif
			ePtr = ptr++;
			int count = --size;
			char* sPtr = symbol;
			while (count--)
				*sPtr++ = *ptr++;
			*sPtr = '\0';
			char* vPtr = NULL;
			if (*ePtr == '$')
			{
				/*	Grab the environment table definition. */

				vPtr = getenv(symbol);
			}
			else if (*ePtr == '%')
			{
				if (symbol[0] != 'm' ||
					symbol[1] != 'a' ||
					symbol[2] != 'c')
				{
					*bPtr++ = *ePtr++;
					ptr = ePtr;
					continue;
				}
				/*	Grab the patch macro value. */

				StringList* value = Macro::findIn(symbol, set);
				if (value)
				{
					value = value->next();
					vPtr = value->contents();
				}
			}
			if (vPtr)
			{
				while (*vPtr)
					*bPtr++ = *vPtr++;
			}
			mustReplace = true;
			continue;
		}
		*bPtr++ = *ptr++;
	}
	*bPtr = '\0';
	return mustReplace;
}

static int usage
(	int errNo)
{
	cout
		<< copyRight << endl << endl
		<<	"Usage:  rochay  [-v]  aFileToPatch  [aPatchScript]  [script args]\n\n"
		<<	"where aFileToPatch names the file to operate on.\n\n"
		<<	"aPatchScript names the file providing the patch commands used\n"
		<<	"in the operation on aFileToPatch.  If omitted, the patch commands\n"
		<<	"are read from standard input without verification.\n\n"
		<<	"-v recites the operations in both passes of the patcher.\n\n"
		<<	"Patching proceeds in two passes: in the first, the entire patch\n"
		<<	"script is tried out to confirm no errors can be detected before\n"
		<<	"starting to update aFileToPatch during the second pass through\n"
		<<	"the script."
		<< endl;
	return errNo;
}

/*
							 ******************
							** Public methods **
							 ******************
*/

void pError
(	const char* txt)
{
		/*	Gripe upon the current anomaly in the system. */

#ifdef __BORLANDC__
	extern char* sys_errlist[];
#else
	extern const char* const sys_errlist[];
#endif
	cout
		<< txt
		<< '('
		<< errno
		<< ") "
		<< (errno > 0 ?
			sys_errlist[errno] :
			"Griping for no apparent reason")
		<< '.'
		<< endl;
}

int main
(	int argc,
	char** argv)
{
		/*	Apply the patch script to the file name given on the CLI. */

	int reply = 0;
	int needReopen = false;
	int forbidUpdate = false;
	Cli theCli = {
		-1, -1, NULL, true, true,
		NULL, 0, NULL, NULL, 0,
		0, 0, 0, 0, 0,
		0, 0, true, 0, NULL,
		0, 0, 0, 0, 0,
		NULL, 0};

	/*	Is the cli at all sane? */

	argc--; argv++;
	if (!argc)
	{
		cout
			<< "No file name to patch was given."
			<< endl;
		return usage(1);
	}
	/*	Check the cli for the -v[erbose] switch. */

	if (**argv == '-' && (*argv)[1] == 'v')
	{
		theCli.verify = true;
		theCli.invert = (*argv)[2] == 'a';
		argc--;
		argv++;
	}
	else if (**argv == '-' && (*argv)[1] == 't')
	{
		/*	Check for 'test' mode. */

		theCli.verify = true;
		theCli.invert = (*argv)[2] == 'a';
		argc--;
		argv++;
		theCli.apply = true;
		theCli.allowCopy = false;
		forbidUpdate = true;
	}
	else if (**argv == '-' && (*argv)[1] == 'a')
	{
		theCli.invert = true;
		argc--;
		argv++;
	}
	/*	Check for output destination override. */

	if (**argv == '-' && (*argv)[1] == 'o')
	{
		theCli.needCopy = false;
		theCli.allowCopy = true;
		argc--;
		argv++;
		if (!argc)
		{
			cout
				<< "No output file name was given."
				<< endl;
			return usage(1);
		}
		theCli.copyName = *argv++;
		argc--;
		needReopen = theCli.haveRedirect = true;
	}
	/*	Check the cli for the name of the file to patch. */

	if (!argc)
	{
		cout
			<< "No file name to patch was given."
			<< endl;
		return usage(1);
	}
	else
	{
		argc--;
		if (**argv == '-')
			needReopen = theCli.haveRedirect = false;
		else
		{
			theCli.fileName = stros1(*argv);
			needReopen = theCli.haveRedirect = true;
		}
		argv++;
	}
	/*	Otherwise, check the cli for the name of the
		patch command file. */

	if (!argc)
	{
		cout << "No patch command file was given." << endl;
		return usage(1);
	}
	theCli.patchName = *argv++;
	argc--;

	/*	Open the command set for processing. */

	theCli.patch = new ifstream(theCli.patchName, ios::nocreate);
	if (!theCli.patch || isFileErrorp(theCli.patch))
	{
		pError(WhoAmI);
		cout
			<< "Unable to open patch command file '"
			<< theCli.patchName
			<< "'."
			<< endl;
		return usage(3);
	}
	cliConfig(theCli, argc, argv);

	char* topFile = theCli.fileName;
	char* topCopy = theCli.copyName;
	reply = prologue(theCli);
	if (!reply)
	{
		theCli.reply = readCmds(theCli, theCli.handle, 0);
		if (theCli.reply)
			return theCli.reply;
		theCli.reply = epilogue(theCli);
		if (theCli.reply)
			return theCli.reply;
		theCli.fileName = topFile;
		theCli.copyName = topCopy;
		theCli.haveRedirect = needReopen;
		section = 0;
		theCli.patchPass++;
		theCli.patch->seekg(0, ios::beg);
		if (prologue(theCli))
			return theCli.reply;
		theCli.patch->clear();
		theCli.patch->seekg(0, ios::beg);
		theCli.reply = readCmds(theCli,
								theCli.modHandle,
								!forbidUpdate && (theCli.allowCopy || theCli.needCopy));
		reply = epilogue(theCli);
	}
	if (*theCli.patch != cin)
		theCli.patch->close();
	cout << "-- Reply is " << reply << endl;
	return reply;
}

StringList* Macro::findIn
(	const char* name,
	Macro* list)
{
		/*	Answer the macro lines for the named macro, limiting
			the search length for %from or %iterate macros. */

	int len;
	isFrom = isEqual(name, "from", 5);
	if (isFrom)
		len = 4;
	else if ((isIterate = isEqual(name, "iterate", 8)))
		len = 7;
	else
		len = strlen(name);
	return list ? 
		list->find(name, len) :
		NULL;
}

void StringList::replace
(	Macro* set)
{
		/*	Replace the contents of the receiver with symbolic values
			from a macro set.  Set errno according to whether or not
			the replacement update was successful. */

	char buff[1024];
	int mustReplace = ::replace(_text, buff, sizeof(buff), set);
	if (mustReplace)
	{
		/*	Update the receiver to a constant string. */

		free(_text);
		errno = 0;
		_text = strdup(buff);
	}
	else
		errno = 0;
}
